#ifndef MEOS_TYPES_TEMPORAL_TEMPORALDURATION_HPP
#define MEOS_TYPES_TEMPORAL_TEMPORALDURATION_HPP

enum class TemporalDuration {
  Temporal,
  Instant,
  InstantSet,
  Sequence,
  SequenceSet,
};

#endif
